uint DEBUG_LEVEL = 0;

type

    Entry_t = struct {
	ulong en_hash;		/* hash function value of name */
	*char en_name;		/* pointer to buffer holding the name */
	ulong en_hunkNumber;
	ulong en_value;
	uint en_nameLen;	/* length in bytes of name - including \e */
    },

    Table_t = struct {
	ulong tb_entryCount;	/* number of actual entries in table */
	ulong tb_usedCount;	/* number of entries in use now */
	ulong tb_shift; 	/* shift to multiply by entryCount */
	ulong tb_mask;		/* mask to modulo by entryCount */
	*Entry_t tb_entries;	/* pointer to actual entry array */
    },

    Size_t = enum {
	sz_32,
	sz_16,
	sz_8,
	sz_common
    },

    Reloc_t = struct {
	*Reloc_t rl_next;
	ulong rl_hunkNumber;
	ulong rl_count; 	/* count of longwords */
	ulong rl_newCount;	/* count of longwords after linking */
	*ulong rl_refOffsets;
	Size_t rl_size;
    },

    Ref_t = struct {
	*Ref_t rf_next;
	*char rf_name;
	ulong rf_count;
	*ulong rf_offsets;
	*Entry_t rf_value;	/* the found definition for the symbol */
	ulong rf_commonLen;
	uint rf_nameLen;
	Size_t rf_size;
    },

    Symbol_t = struct {
	*Symbol_t sy_next;
	*char sy_name;
	ulong sy_value;
	uint sy_nameLen;
    },

    Hunk_t = struct {
	*Hunk_t h_next;
	ulong h_type;
	ulong h_number;
	ulong h_length; 	/* length in bytes, not longwords */
	arbptr h_buffer;
	*Reloc_t h_relocs;
	*Ref_t h_refs;
	Table_t h_symbols;
	Table_t h_externs;
	ulong h_smallOffset;	/* used when smallXXX in effect */
	*Symbol_t h_symbolic;	/* symbolic info */
	*byte h_debug;		/* debug info */
	ulong h_debugLen;
    },

    File_t = struct {
	*File_t fl_next;
	*char fl_name;
	Handle_t fl_handle;
	ulong fl_fileNumber;
	*Hunk_t fl_hunkList;
    };

uint
    EN_EMPTY = 0,		/* slot never been used */
    EN_FREE = 0xffff;		/* slot now empty */

extern

    /* DRLink.d */

    print(*char st)void,
    printNumber(ulong n)void,
    freeHunkList(*Hunk_t h)void,
    freeAll()void,
    abortTail()void,
    abort(*char message)void,
    abortString(*char m1, m2, m3)void,
    abortInput(*char message)void,
    printInputName()void,
    getMem(ulong length)arbptr,
    setInput(Handle_t fd; *char name)void,
    unsetInput()Handle_t,
    allowEof(bool flag)void,
    hadEof()bool,
    doRead(arbptr buffer; ulong len)void,
    readLong()ulong,
    skipLongs(ulong count)void,
    doWrite(arbptr buffer; ulong len)void,
    writeLong(ulong l)void,
    writeName(*char name; uint nameLen)void,

    /* process.d */

    processInit()void,
    processTerm()void,
    readFrom(bool define)bool,
    addLib(*char fileName)void,
    processLibraries(bool define)bool,
    checkUnDef()bool,
    writeOutputFile(bool smallCode, smallData, smallBss, define)void,

    /* symbol.d */

    symCreate(*Table_t tb; ulong shift)void,
    symDestroy(*Table_t tb)void,
    symFlush(*Table_t tb)void,
    symEmpty(*Table_t tb)void,
    symEnter(*Table_t tb; *char name; uint nameLen)*Entry_t,
    symLookup(*Table_t tb; *char name; uint nameLen)*Entry_t,
    symDelete(*Table_t tb; *char name; uint nameLen)void,
    symCopy(*Table_t oldTable, newTable)void;
