#drinc:exec/miscellaneous.g
#drinc:exec/memory.g
#drinc:exec/tasks.g
#drinc:libraries/dos.g
#drinc:util.g
#DRLink.g

char
    VERSION_1 = '1',
    VERSION_2 = '0',
    DATE = '1';

ulong
    BLOCK_SIZE = 512,
    INPUT_BUFFER_SIZE = BLOCK_SIZE * 8,
    OUTPUT_BUFFER_SIZE = BLOCK_SIZE * 8;

*char WithBuffer, WithPointer;
ulong WithLen;

Handle_t InputFd;
Handle_t OutputFd;
Handle_t StdOut;
*char InputName;
bool AllowEof, HadEof;

ulong InputPos, InputMax, OutputPos;
[INPUT_BUFFER_SIZE] byte InputBuffer;
[OUTPUT_BUFFER_SIZE] byte OutputBuffer;

*char FromString, LibraryString, ToString;
bool SmallCode, SmallData, SmallBss, Define;

proc print(*char st)void:
    register *char p;

    p := st;
    while p* ~= '\e' do
	p := p + sizeof(char);
    od;
    ignore Write(StdOut, st, p - st);
corp;

proc printNumber(register ulong n)void:
    [11] char buffer;
    register *char p;

    p := &buffer[10];
    p* := '\e';
    while
	p := p - sizeof(char);
	p* := n % 10 + '0';
	n := n / 10;
	n ~= 0
    do
    od;
    print(p);
corp;

#A
proc freeHunkList(register *Hunk_t h)void:
    register *Hunk_t hTemp;
    register *Reloc_t rl;
    register *Ref_t rf @ rl;
    register *Symbol_t sy @ rl;
    register proc(arbptr p; ulong len)void r_FreeMem;

    r_FreeMem := FreeMem;
    while h ~= nil do
	hTemp := h;
	h := h*.h_next;
	if DEBUG_LEVEL > 2 then
	    print("freeing hunk ");
	    printNumber(hTemp*.h_number);
	    print("\n");
	fi;
	if hTemp*.h_debug ~= nil then
	    if DEBUG_LEVEL > 3 then
		print("freeing debug info\n");
	    fi;
	    r_FreeMem(hTemp*.h_debug, hTemp*.h_debugLen);
	fi;
	while
	    sy := hTemp*.h_symbolic;
	    sy ~= nil
	do
	    if sy*.sy_name ~= nil then
		r_FreeMem(sy*.sy_name, sy*.sy_nameLen);
	    fi;
	    hTemp*.h_symbolic := sy*.sy_next;
	    if DEBUG_LEVEL > 3 then
		print("freeing symbolic piece\n");
	    fi;
	    r_FreeMem(sy, sizeof(Symbol_t));
	od;
	symDestroy(&hTemp*.h_externs);
	symDestroy(&hTemp*.h_symbols);
	while
	    rf := hTemp*.h_refs;
	    rf ~= nil
	do
	    if rf*.rf_offsets ~= nil then
		if DEBUG_LEVEL > 3 then
		    print("freeing ref offsets\n");
		fi;
		r_FreeMem(rf*.rf_offsets, rf*.rf_count * 4);
	    fi;
	    if rf*.rf_name ~= nil then
		r_FreeMem(rf*.rf_name, rf*.rf_nameLen);
	    fi;
	    hTemp*.h_refs := rf*.rf_next;
	    if DEBUG_LEVEL > 3 then
		print("freeing ref\n");
	    fi;
	    r_FreeMem(rf, sizeof(Ref_t));
	od;
	while
	    rl := hTemp*.h_relocs;
	    rl ~= nil
	do
	    if rl*.rl_refOffsets ~= nil then
		if DEBUG_LEVEL > 3 then
		    print("freeing reloc ref offsets\n");
		fi;
		r_FreeMem(rl*.rl_refOffsets, rl*.rl_count * 4);
	    fi;
	    hTemp*.h_relocs := rl*.rl_next;
	    if DEBUG_LEVEL > 3 then
		print("freeing reloc\n");
	    fi;
	    r_FreeMem(rl, sizeof(Reloc_t));
	od;
	if hTemp*.h_buffer ~= nil then
	    if DEBUG_LEVEL > 3 then
		print("freeing hunk data\n");
	    fi;
	    if hTemp*.h_length ~= 0 then
		r_FreeMem(hTemp*.h_buffer, hTemp*.h_length);
	    fi;
	fi;
	if DEBUG_LEVEL > 3 then
	    print("freeing hunk\n");
	fi;
	r_FreeMem(hTemp, sizeof(Hunk_t));
    od;
corp;

proc freeAll()void:

    if OutputFd ~= 0 then
	Close(OutputFd);
    fi;
    if InputFd ~= 0 then
	Close(InputFd);
    fi;
    processTerm();
    /* do this last, since it can be referenced by the library list */
    if WithBuffer ~= nil then
	if DEBUG_LEVEL > 3 then
	    print("freeing WithBuffer\n");
	fi;
	FreeMem(WithBuffer, WithLen + 1);
    fi;
corp;

proc abortTail()void:

    print(" - aborting\n");
    freeAll();
    Exit(RETURN_FAIL);
corp;

proc abort(*char message)void:

    print("DRLink: ");
    print(message);
    abortTail();
corp;

proc abortString(*char m1, m2, m3)void:

    print("DRLink: ");
    print(m1);
    print(m2);
    print(m3);
    abortTail();
corp;

proc abortInput(*char m1)void:

    print("DRLink: ");
    print(m1);
    print(" '");
    print(InputName);
    print("'");
    abortTail();
corp;

proc printInputName()void:

    print(InputName);
corp;

proc getMem(ulong length)arbptr:
    arbptr p;

    p := AllocMem(length, 0);
    if p = nil then
	abort("out of memory");
    fi;
    p
corp;

proc setInput(Handle_t fd; *char name)void:

    InputName := name;
    if fd = 0 then
	InputFd := Open(name, MODE_READONLY);
	if InputFd = 0 then
	    abortString("cannot open library '", name, "'");
	fi;
    else
	InputFd := fd;
	ignore Seek(fd, 0, OFFSET_BEGINNING);
    fi;
    InputPos := 0;
    InputMax := 0;
    HadEof := false;
corp;

proc unsetInput()Handle_t:
    Handle_t fd;

    fd := InputFd;
    InputFd := 0;
    fd
corp;

proc allowEof(bool flag)void:

    AllowEof := flag;
corp;

proc hadEof()bool:

    HadEof
corp;

proc doRead(arbptr buffer; register ulong len)void:
    register ulong count;
    register *byte p, q;
    register long got, want;

    p := buffer;
    count := InputPos;
    q := &InputBuffer[count];
    count := InputMax - count;
    while len ~= 0 and count ~= 0 do
	p* := q*;
	p := p + 1;
	q := q + 1;
	count := count - 1;
	len := len - 1;
    od;
    InputPos := InputMax - count;
    if len = 0 then
	return;
    fi;
    if len >= BLOCK_SIZE then
	if SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C ~= 0 then
	    print("Interrupt! - DRLink exiting.\n");
	    freeAll();
	    Exit(RETURN_WARN);
	fi;
	want := len / BLOCK_SIZE * BLOCK_SIZE;
	got := Read(InputFd, p, want);
	if got = 0 and AllowEof then
	    HadEof := true;
	    return;
	fi;
	if got ~= want then
	    abortInput("bad read from");
	fi;
	len := len - want;
	if len = 0 then
	    return;
	fi;
	p := p + want;
    fi;
    q := &InputBuffer[0];
    if SetSignal(0, SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C ~= 0 then
	print("Interrupt! - DRLink exiting.\n");
	freeAll();
	Exit(RETURN_WARN);
    fi;
    got := Read(InputFd, q, INPUT_BUFFER_SIZE);
    if got = 0 and AllowEof then
	HadEof := true;
	return;
    fi;
    if got <= 0 then
	abortInput("bad read from");
    fi;
    InputMax := got;
    InputPos := len;
    while len ~= 0 do
	p* := q*;
	p := p + 1;
	q := q + 1;
	len := len - 1;
    od;
corp;

proc readLong()ulong:
    ulong w;

    doRead(&w, 4);
    w
corp;

proc skipLongs(ulong count)void:
    [512] byte buffer;
    register ulong len;

    while count ~= 0 do
	if count >= 512 / 4 then
	    count := count - 512 / 4;
	    len := 512;
	else
	    len := count * 4;
	    count := 0;
	fi;
	doRead(&buffer[0], len);
    od;
corp;

proc addFrom(*char fileName)bool:
    bool ok;

    InputName := fileName;
    if DEBUG_LEVEL > 0 then
	print("reading 'from' file '");
	print(fileName);
	print("'\n");
    fi;
    InputName := fileName;
    InputFd := Open(fileName, MODE_READONLY);
    if InputFd = 0 then
	abortInput("can't open 'from' file");
    fi;
    HadEof := false;
    InputPos := 0;
    InputMax := 0;
    ok := readFrom(Define);
    Close(InputFd);
    InputFd := 0;
    ok
corp;

proc processFiles()bool:
    register *char p, start;
    bool ok;

    ok := true;
    p := FromString;
    if p ~= nil then
	while p* ~= '\e' do
	    start := p;
	    while p* ~= '\e' and p* ~= '+' do
		p := p + sizeof(char);
	    od;
	    if p* ~= '\e' then
		p* := '\e';
		p := p + sizeof(char);
	    fi;
	    if not addFrom(start) then
		ok := false;
	    fi;
	od;
    fi;
    p := LibraryString;
    if p ~= nil then
	while p* ~= '\e' do
	    start := p;
	    while p* ~= '\e' and p* ~= '+' do
		p := p + sizeof(char);
	    od;
	    if p* ~= '\e' then
		p* := '\e';
		p := p + sizeof(char);
	    fi;
	    addLib(start);
	od;
    fi;
    ok
corp;

proc openOutput()void:

    OutputFd := Open(ToString, MODE_NEWFILE);
    if OutputFd = 0 then
	abortString("can't open/create output file '", ToString, "'");
    fi;
    OutputPos := 0;
corp;

proc flushOutput()void:

    if Write(OutputFd, &OutputBuffer[0], OutputPos) ~= OutputPos then
	abortString("error on write to output file '", ToString, "'");
    fi;
corp;

proc doWrite(arbptr buffer; register ulong len)void:
    register *byte p, q;
    register ulong count;

    count := OutputPos;
    p := &OutputBuffer[count];
    count := OUTPUT_BUFFER_SIZE - count;
    q := buffer;
    while len ~= 0 do
	if count = 0 then
	    OutputPos := OUTPUT_BUFFER_SIZE;
	    flushOutput();
	    if len >= BLOCK_SIZE then
		count := len / BLOCK_SIZE * BLOCK_SIZE;
		if Write(OutputFd, q, count) ~= count then
		    abortString("error on write to output file '",
				ToString, "'");
		fi;
		len := len - count;
		q := q + count;
	    fi;
	    count := OUTPUT_BUFFER_SIZE;
	    p := &OutputBuffer[0];
	else
	    p* := q*;
	    p := p + 1;
	    q := q + 1;
	    len := len - 1;
	    count := count - 1;
	fi;
    od;
    OutputPos := OUTPUT_BUFFER_SIZE - count;
corp;

proc writeLong(ulong l)void:

    doWrite(&l, 4);
corp;

proc writeName(*char name; uint nameLen)void:
    ulong len;

    len := (nameLen + 3) / 4;
    writeLong(len);
    doWrite(name, nameLen);
    len := len * 4;
    while len ~= nameLen do
	len := len - 1;
	doWrite("", 1);
    od;
corp;

proc handleOtherParameter(register *char par; proc()*char getWord)void:

    if CharsEqual(par, "from") then
	par := getWord();
	if par = nil then
	    abort("missing filename after 'from'");
	fi;
	FromString := par;
    elif CharsEqual(par, "lib") or CharsEqual(par, "library") then
	par := getWord();
	if par = nil then
	    abort("missing filename after 'library' or 'lib'");
	fi;
	LibraryString := par;
    elif CharsEqual(par, "to") then
	par := getWord();
	if par = nil then
	    abort("missing filename after 'to'");
	fi;
	ToString := par;
    elif CharsEqual(par, "smallcode") or CharsEqual(par, "sc") then
	SmallCode := true;
    elif CharsEqual(par, "smalldata") or CharsEqual(par, "sd") then
	SmallData := true;
    elif CharsEqual(par, "smallbss") or CharsEqual(par, "sb") then
	SmallBss := true;
    elif CharsEqual(par, "define") or CharsEqual(par, "def") then
	Define := true;
    else
	abortString("invalid parameter '", par, "'");
    fi;
corp;

proc getWithWord()*char:
    register *char start;

    start := WithPointer;
    if start* = '\e' then
	return(nil);
    fi;
    while WithPointer* ~= ' ' and WithPointer* ~= '\e' do
	WithPointer := WithPointer + sizeof(char);
    od;
    if WithPointer* ~= '\e' then
	WithPointer* := '\e';
	WithPointer := WithPointer + sizeof(char);
    fi;
    start
corp;

proc parseWithFile()void:
    register *char p, q;

    /* Run through the 'with' file contents and compress out whitespace
       and *-newline pairs. Also turn all non-* newlines into a space. */
    p := WithBuffer;
    while p* ~= '\e' do
	case p*
	incase ' ':
	incase '\t':
	    q := p + sizeof(char);
	    if q* = ' ' or q* = '\t' then
		CharsCopy(p, q);
		p* := ' ';
	    else
		p := q;
	    fi;
	incase '*':
	    q := p + sizeof(char);
	    if q* = '\n' then
		CharsCopy(p, q + sizeof(char));
	    else
		p := q;
	    fi;
	incase '\n':
	    p* := ' ';
	    /* loop around again and handle it above */
	default:
	    p := p + sizeof(char);
	esac;
    od;
    WithPointer := WithBuffer;
    if WithPointer* = ' ' then
	WithPointer := WithPointer + sizeof(char);
    fi;
    while
	p := getWithWord();
	p ~= nil
    do
	handleOtherParameter(p, getWithWord);
    od;
corp;

proc main()void:
    extern _d_pars_initialize()void;
    char COPYRIGHT = '\(0xa9)';
    register *char par;

    if OpenExecLibrary(0) ~= nil then
	if OpenDosLibrary(0) ~= nil then
	    _d_pars_initialize();
	    StdOut := Output();
	    print("DRLink V\(VERSION_1).\(VERSION_2) \(COPYRIGHT) 199\(DATE) "
		  "by Chris Gray\n");
	    AllowEof := false;
	    WithBuffer := nil;
	    FromString := nil;
	    LibraryString := nil;
	    ToString := nil;
	    SmallCode := false;
	    SmallData := false;
	    SmallBss := false;
	    Define := false;
	    InputFd := 0;
	    OutputFd := 0;
	    processInit();
	    while
		par := GetPar();
		par ~= nil
	    do
		if CharsEqual(par, "with") then
		    par := GetPar();
		    if par = nil then
			abort("missing file name after 'with'");
		    fi;
		    InputFd := Open(par, MODE_READONLY);
		    if InputFd = 0 then
			abortString("can't open 'with' file '", par, "'");
		    fi;
		    ignore Seek(InputFd, 0, OFFSET_END);
		    WithLen := Seek(InputFd, 0, OFFSET_BEGINNING);
		    WithBuffer := getMem(WithLen + 1);
		    if WithBuffer = nil then
			abortString("can't buffer 'with' file '", par, "'");
		    fi;
		    if Read(InputFd, WithBuffer, WithLen) ~= WithLen then
			abortString("bad read from 'with' file '", par, "'");
		    fi;
		    Close(InputFd);
		    InputFd := 0;
		    (WithBuffer + WithLen)* := '\e';
		    parseWithFile();
		else
		    handleOtherParameter(par, GetPar);
		fi;
	    od;
	    if ToString = nil then
		abort("no output name specified");
	    fi;
	    DeleteFile(ToString);
	    if processFiles() then
		if processLibraries(Define) then
		    if checkUnDef() then
			openOutput();
			writeOutputFile(SmallCode, SmallData, SmallBss,Define);
			flushOutput();
		    fi;
		fi;
	    fi;
	    freeAll();
	    CloseDosLibrary();
	fi;
	CloseExecLibrary();
    fi;
corp;
