#drinc:exec/memory.g
#drinc:libraries/dos.g
#drinc:util.g
#DRLink.g

/*
 * One of the most annoying problems I've had getting this working is that of
 * the zero-length code hunk that 'assem' insists on putting at the beginning
 * of 'drlib:drstartf.o'. I can't just throw it away on input, since that
 * messes up inter-hunk references done via hunk number in that file. I also
 * can't just emit it normally if 'smallcode' is not specified, since then
 * it becomes the entry point of the program, and, since scatter-loading is
 * happening, whatever happens to be after it in memory is the true entry
 * point. Trying to delete it sometime during 'writeOutputFile' can mess up
 * the inter-hunk referencing done by hunk number there. So, I do what I
 * suspect is what BLink does - stick in a 'jmp' to the first real code hunk.
 */

ulong
    NAME_MAX = 255 * 4,
    DEF_MAP_SIZE = 1024;

ulong
    HUNK_UNIT = 999,
    HUNK_NAME = 1000,
    HUNK_CODE = 1001,
    HUNK_DATA = 1002,
    HUNK_BSS = 1003,
    HUNK_RELOC32 = 1004,
    HUNK_RELOC16 = 1005,
    HUNK_RELOC8 = 1006,
    HUNK_EXT = 1007,
    HUNK_SYMBOL = 1008,
    HUNK_DEBUG = 1009,
    HUNK_END = 1010,
    HUNK_HEADER = 1011,

    EXT_DEF = 1,
    EXT_REF32 = 129,
    EXT_COMMON = 130,
    EXT_REF16 = 131,
    EXT_REF8 = 132,

    VAR_ARRAY = 1,

    HUNK_MASK = 0x00ffffff;

Table_t GlobalDefTable, GlobalExtTable;
Table_t LocalDefTable, LocalExtTable;

ulong HunkNumber, FirstHunk;
*Hunk_t HunkList;
**Hunk_t HunkTail;
*[VAR_ARRAY]*Hunk_t HunkMap;
ulong HunkMapSize;
*File_t LibraryList;
bool HadError, FirstPass;

[NAME_MAX + 1] char NameBuffer;
uint NameLen;

uint MAX_HUNKS = 1000;
[MAX_HUNKS] bool HunksUsed;

bool ForceSmallCode, ForceSmallData, ForceSmallBss;

proc processInit()void:

    HunkList := nil;
    HunkTail := &HunkList;
    LibraryList := nil;
    HunkMap := nil;
    HunkMapSize := DEF_MAP_SIZE;
    HunkMap := getMem(sizeof(*Hunk_t) * DEF_MAP_SIZE);
    HunkNumber := 0;
    HadError := false;
    ForceSmallCode := false;
    ForceSmallData := false;
    ForceSmallBss := false;
    GlobalDefTable.tb_usedCount := 0;
    GlobalExtTable.tb_usedCount := 0;
    LocalDefTable.tb_usedCount := 0;
    LocalDefTable.tb_usedCount := 0;
    symCreate(&GlobalDefTable, 11);
    symCreate(&GlobalExtTable, 9);
    symCreate(&LocalDefTable, 8);
    symCreate(&LocalExtTable, 8);
corp;

proc processTerm()void:
    register *File_t fl;

    freeHunkList(HunkList);
    while
	fl := LibraryList;
	fl ~= nil
    do
	if fl*.fl_handle ~= 0 then
	    Close(fl*.fl_handle);
	fi;
	LibraryList := fl*.fl_next;
	if DEBUG_LEVEL > 3 then
	    print("freeing file list\n");
	fi;
	/* only anything there when we abort during library processing */
	freeHunkList(fl*.fl_hunkList);
	FreeMem(fl, sizeof(File_t));
    od;
    if DEBUG_LEVEL > 3 then
	print("freeing HunkMap\n");
    fi;
    FreeMem(HunkMap, sizeof(*Hunk_t) * HunkMapSize);
    if DEBUG_LEVEL > 1 then
	print("destroying LocalExtTable\n");
    fi;
    symDestroy(&LocalExtTable);
    if DEBUG_LEVEL > 1 then
	print("destroying LocalDefTable\n");
    fi;
    symDestroy(&LocalDefTable);
    if DEBUG_LEVEL > 1 then
	print("destroying GlobalExtTable\n");
    fi;
    symDestroy(&GlobalExtTable);
    if DEBUG_LEVEL > 1 then
	print("destroying GlobalDefTable\n");
    fi;
    symDestroy(&GlobalDefTable);
corp;

proc saveName(**char pCopy; *uint pLen)void:

    pCopy* := getMem(NameLen);
    BlockCopy(pCopy*, &NameBuffer[0], NameLen);
    pLen* := NameLen;
corp;

proc expandHunkMap()void:
    ulong newSize;
    *[VAR_ARRAY]*Hunk_t newMap;

    /* be carefull, so that if 'getMem' aborts, all is well. */
    newSize := HunkMapSize * 2;
    newMap := getMem(sizeof(* Hunk_t) * newSize);
    BlockCopy(newMap, HunkMap, sizeof(*Hunk_t) * HunkMapSize);
    FreeMem(HunkMap, sizeof(* Hunk_t) * HunkMapSize);
    HunkMapSize := newSize;
    HunkMap := newMap;
corp;

proc readName(register ulong count)void:
    register uint i, j;
    register char ch;
    ulong w;
    [4]char wChar @ w;

    if count > NAME_MAX then
	abortInput("symbol too long in");
    fi;
    j := 0;
    while count ~= 0 do
	w := readLong();
	i := 0;
	while i ~= 4 do
	    ch := wChar[i];
	    if ch = '\e' then
		i := 4;
	    else
		NameBuffer[j] := ch;
		j := j + 1;
		i := i + 1;
	    fi;
	od;
	count := count - 1;
    od;
    NameBuffer[j] := '\e';
    j := j + 1;
    NameLen := j;
corp;

proc readSymbols(register *Hunk_t h)void:
    register *Symbol_t sy;
    register **Symbol_t syTail;
    register ulong w;

    syTail := &h*.h_symbolic;
    while syTail* ~= nil do
	syTail := &syTail**.sy_next;
    od;
    while
	w := readLong();
	w ~= 0
    do
	readName(w & 0x00ffffff);
	sy := getMem(sizeof(Symbol_t));
	sy*.sy_name := nil;
	sy*.sy_next := nil;
	syTail* := sy;
	syTail := &sy*.sy_next;
	saveName(&sy*.sy_name, &sy*.sy_nameLen);
	sy*.sy_value := readLong();
    od;
corp;

proc readHunk(register ulong hunkType; **Hunk_t pHunk;
	      bool inLibrary, define)void:
    *Table_t defTable, extTable;
    register *Hunk_t h;
    register *Reloc_t rl;
    register *Ref_t rf;
    register *Entry_t en @ rl;
    register ulong w;
    register byte extKind;

    if FirstPass then
	if HunkNumber >= HunkMapSize then
	    expandHunkMap();
	fi;
	HunkMap*[HunkNumber] := nil;
    fi;
    h := getMem(sizeof(Hunk_t));
    h*.h_next := nil;
    h*.h_number := HunkNumber;
    h*.h_symbols.tb_entryCount := 0;
    h*.h_externs.tb_entryCount := 0;
    h*.h_symbolic := nil;
    h*.h_debug := nil;
    if inLibrary then
	defTable := &LocalDefTable;
	extTable := &LocalExtTable;
    else
	defTable := &GlobalDefTable;
	extTable := &GlobalExtTable;
    fi;
    if hunkType = HUNK_SYMBOL then
	readSymbols(h);
	hunkType := readLong();
    elif hunkType = HUNK_DEBUG then
	h*.h_debugLen := readLong() * 4;
	if h*.h_debugLen ~= 0 then
	    h*.h_debug := getMem(h*.h_debugLen);
	    doRead(h*.h_debug, h*.h_debugLen);
	fi;
	hunkType := readLong();
    fi;
    w := hunkType & HUNK_MASK;
    if w ~= HUNK_CODE and w ~= HUNK_DATA and w ~= HUNK_BSS then
	FreeMem(h, sizeof(Hunk_t));
	print("DRLink: unsupported hunk type (");
	printNumber(hunkType);
	print(") in '");
	printInputName();
	print("'");
	abortTail();
    fi;
    h*.h_type := hunkType;	/* NOT w - want the fast/chip bits */
    h*.h_buffer := nil;
    h*.h_relocs := nil;
    h*.h_refs := nil;
    h*.h_length := 0;
    /* from here on, we must maintain the hunk in a valid state, since we
       can abort at any instant */
    pHunk* := h;
    if DEBUG_LEVEL > 3 then
	print("hunk ");
	printNumber(h*.h_number);
    fi;
    h*.h_length := readLong() * 4;
    if w = HUNK_BSS then
	if DEBUG_LEVEL > 3 then
	    print(" bss hunk, length ");
	    printNumber(h*.h_length);
	    print("\n");
	fi;
    else
	if DEBUG_LEVEL > 3 then
	    if w = HUNK_CODE then
		print(" text");
	    else
		print(" data");
	    fi;
	    print(" hunk, length ");
	    printNumber(h*.h_length);
	    print("\n");
	fi;
	if h*.h_length ~= 0 then
	    h*.h_buffer := getMem(h*.h_length);
	    doRead(h*.h_buffer, h*.h_length);
	fi;
    fi;
    hunkType := readLong();
    while hunkType = HUNK_RELOC32 or hunkType = HUNK_RELOC16 or
	hunkType = HUNK_RELOC8 or
	hunkType = HUNK_EXT or hunkType = HUNK_SYMBOL or hunkType = HUNK_DEBUG
    do
	if hunkType = HUNK_RELOC32 or hunkType = HUNK_RELOC16 or
	    hunkType = HUNK_RELOC8
	then
	    while
		w := readLong();
		w ~= 0
	    do
		rl := getMem(sizeof(Reloc_t));
		rl*.rl_count := w;
		rl*.rl_newCount := w;
		rl*.rl_refOffsets := nil;
		if hunkType = HUNK_RELOC32 then
		    rl*.rl_size := sz_32;
		else
		    w := h*.h_type & HUNK_MASK;
		    if w = HUNK_CODE then
			ForceSmallCode := true;
		    elif w = HUNK_DATA then
			ForceSmallData := true;
		    else
			ForceSmallBss := true;
		    fi;
		    if hunkType = HUNK_RELOC16 then
			rl*.rl_size := sz_16;
		    else
			rl*.rl_size := sz_8;
		    fi;
		fi;
		rl*.rl_next := h*.h_relocs;
		h*.h_relocs := rl;
		/* either of the following lines can abort */
		rl*.rl_hunkNumber := readLong() + FirstHunk;
		rl*.rl_refOffsets := getMem(w * 4);
		doRead(rl*.rl_refOffsets, w * 4);
		if DEBUG_LEVEL > 3 then
		    print("reloc to hunk ");
		    printNumber(rl*.rl_hunkNumber);
		    print("\n");
		fi;
	    od;
	elif hunkType = HUNK_EXT then
	    while
		w := readLong();
		w ~= 0
	    do
		extKind := w >> 24;
		w := w & 0xffffff;
		if extKind = EXT_DEF then
		    readName(w);
		    if DEBUG_LEVEL > 3 then
			print("defining '");
			print(&NameBuffer[0]);
			print("'\n");
		    fi;
		    w := readLong();
		    if symLookup(defTable, &NameBuffer[0], NameLen) ~= nil then
			print("DRLink: ");
			printInputName();
			print(" - '");
			print(&NameBuffer[0]);
			print("' is multiply defined\n");
			HadError := true;
		    else
			en := symEnter(defTable, &NameBuffer[0], NameLen);
			en*.en_hunkNumber := h*.h_number;
			en*.en_value := w;
			/* see if it satisfies an existing extref in a .r */
			if not inLibrary then
			    symDelete(&GlobalExtTable,
				      en*.en_name, en*.en_nameLen);
			    if define then
				en := symEnter(&LocalDefTable,
					       &NameBuffer[0], NameLen);
				en*.en_value := w;
			    fi;
			fi;
		    fi;
		elif extKind = EXT_REF32 or extKind = EXT_COMMON or
		    extKind = EXT_REF16 or extKind = EXT_REF8
		then
		    readName(w);
		    if DEBUG_LEVEL > 3 then
			print("ref to '");
			print(&NameBuffer[0]);
			print("'\n");
		    fi;
		    rf := getMem(sizeof(Ref_t));
		    rf*.rf_offsets := nil;
		    rf*.rf_name := nil;
		    rf*.rf_value := nil;
		    if extKind = EXT_REF32 then
			rf*.rf_size := sz_32;
		    elif extKind = EXT_REF16 then
			rf*.rf_size := sz_16;
		    elif extKind = EXT_REF8 then
			rf*.rf_size := sz_8;
		    else
			rf*.rf_size := sz_common;
		    fi;
		    rf*.rf_next := h*.h_refs;
		    h*.h_refs := rf;
		    saveName(&rf*.rf_name, &rf*.rf_nameLen);
		    if extKind = EXT_COMMON then
			rf*.rf_commonLen := readLong();
		    fi;
		    rf*.rf_count := readLong();
		    rf*.rf_offsets := getMem(rf*.rf_count * 4);
		    doRead(rf*.rf_offsets, rf*.rf_count * 4);
		    if not inLibrary then
			en := symLookup(defTable, rf*.rf_name, rf*.rf_nameLen);
		    else
			en := nil;
		    fi;
		    if en = nil then
			/* either we are loading a library hunk, or the
			   symbol was not already defined */
			ignore symEnter(extTable, rf*.rf_name, rf*.rf_nameLen);
		    fi;
		else
		    print("DRLink: unsupported EXT_type (");
		    printNumber(extKind);
		    print(") in '");
		    printInputName();
		    print("'");
		    abortTail();
		fi;
	    od;
	elif hunkType = HUNK_SYMBOL then
	    readSymbols(h);
	else
	    if h*.h_debug = nil then
		h*.h_debugLen := readLong() * 4;
		if h*.h_debugLen ~= 0 then
		    h*.h_debug := getMem(h*.h_debugLen);
		    doRead(h*.h_debug, h*.h_debugLen);
		fi;
	    else
		/* hmm, what am I supposed to do, concatenate them? */
		skipLongs(readLong());
	    fi;
	fi;
	hunkType := readLong();
    od;
    if hunkType ~= HUNK_END then
	abortInput("missing HUNK_END in");
    fi;
    HunkNumber := HunkNumber + 1;
    if inLibrary or define then
	symCopy(&LocalDefTable, &h*.h_symbols);
	symFlush(&LocalDefTable);
    fi;
    if inLibrary then
	symCopy(&LocalExtTable, &h*.h_externs);
	symFlush(&LocalExtTable);
    fi;
corp;

proc readRealHunk(**Hunk_t pHunk; bool inLibrary, define)bool:
    ulong hunkType;

    while
	allowEof(true);
	hunkType := readLong();
	allowEof(false);
	if hadEof() then
	    return(false);
	fi;
	if hunkType = HUNK_UNIT then
	    /* skip this one */
	    skipLongs(readLong());
	    FirstHunk := HunkNumber;
	    hunkType := readLong();
	fi;
	if hunkType = HUNK_NAME then
	    /* skip this one too */
	    skipLongs(readLong());
	    hunkType := readLong();
	fi;
	hunkType = HUNK_END
    do
    od;
    readHunk(hunkType, pHunk, inLibrary, define);
    true
corp;

proc readFrom(bool define)bool:
    register *Hunk_t h;

    FirstHunk := HunkNumber;
    while readRealHunk(HunkTail, false, define) do
	h := HunkTail*;
	HunkTail := &h*.h_next;
	HunkMap*[HunkNumber - 1] := h;
    od;
    not HadError
corp;

proc addLib(*char fileName)void:
    register *File_t fl;
    register **File_t flp;

    fl := getMem(sizeof(File_t));
    fl*.fl_next := nil;
    fl*.fl_name := fileName;
    fl*.fl_handle := 0;
    fl*.fl_hunkList := nil;
    flp := &LibraryList;
    while flp* ~= nil do
	flp := &flp**.fl_next;
    od;
    flp* := fl;
corp;

proc addHunkFromLibrary(*File_t fl; register *Hunk_t h; bool define)void:
    register *Entry_t en, en2;
    register ulong count;

    if DEBUG_LEVEL > 2 then
	print("using hunk ");
	printNumber(h*.h_number);
	print("\n");
    fi;
    /* we need the hunk - move it to global lists */
    HunkMap*[h*.h_number] := h;
    HunkTail* := h;
    HunkTail := &h*.h_next;
    h*.h_next := nil;
    /* merge in any new external references, after checking for defined */
    if h*.h_externs.tb_usedCount ~= 0 then
	en := h*.h_externs.tb_entries;
	count := h*.h_externs.tb_entryCount;
	while count ~= 0 do
	    count := count - 1;
	    if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
		if symLookup(&GlobalDefTable,
			     en*.en_name, en*.en_nameLen) = nil and
		    symLookup(&GlobalExtTable,
			      en*.en_name, en*.en_nameLen) = nil
		then
		    ignore symEnter(&GlobalExtTable,
				    en*.en_name, en*.en_nameLen);
		fi;
	    fi;
	    en := en + sizeof(Entry_t);
	od;
    fi;
    symDestroy(&h*.h_externs);
    /* all symbols defined here - move to the global
       symbol list and delete from extref list if there */
    en := h*.h_symbols.tb_entries;
    count := h*.h_symbols.tb_entryCount;
    while count ~= 0 do
	count := count - 1;
	if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
	    if symLookup(&GlobalDefTable, en*.en_name, en*.en_nameLen) ~= nil
	    then
		/* symbol already defined - problem! */
		print("DRLink: ");
		print(fl*.fl_name);
		print(" - '");
		print(en*.en_name);
		print("' is multiply defined\n");
		HadError := true;
	    else
		en2 := symEnter(&GlobalDefTable, en*.en_name, en*.en_nameLen);
		en2*.en_hunkNumber := en*.en_hunkNumber;
		en2*.en_value := en*.en_value;
		symDelete(&GlobalExtTable, en*.en_name, en*.en_nameLen);
	    fi;
	fi;
	en := en + sizeof(Entry_t);
    od;
    if not define then
	symDestroy(&h*.h_symbols);
    fi;
corp;

proc selectHunk(*File_t fl; *Hunk_t h; register ulong beginHunk)void:
    register *Reloc_t rl;
    register *Entry_t en @ rl;
    register *Hunk_t h2;
    register ulong count;

    if DEBUG_LEVEL > 2 then
	print("processing selected hunk ");
	printNumber(h*.h_number);
	print("\n");
    fi;
    /* which hunks does this hunk reference by reloc? */
    rl := h*.h_relocs;
    while rl ~= nil do
	if DEBUG_LEVEL > 1 then
	    print("reloc to hunk ");
	    printNumber(rl*.rl_hunkNumber);
	    print("\n");
	fi;
	HunksUsed[rl*.rl_hunkNumber - beginHunk] := true;
	rl := rl*.rl_next;
    od;
    /* which hunks in this library does it reference by name? */
    en := h*.h_externs.tb_entries;
    count := h*.h_externs.tb_entryCount;
    while count ~= 0 do
	count := count - 1;
	if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
	    h2 := fl*.fl_hunkList;
	    while h2 ~= nil do
		if symLookup(&h2*.h_symbols,en*.en_name,en*.en_nameLen) ~= nil
		    and
		    symLookup(&GlobalDefTable,en*.en_name,en*.en_nameLen) = nil
		then
		    if DEBUG_LEVEL > 1 then
			if not HunksUsed[h2*.h_number - beginHunk] then
			    print("in-library ref via '");
			    print(en*.en_name);
			    print("' to hunk ");
			    printNumber(h2*.h_number);
			    print("\n");
			fi;
		    fi;
		    HunksUsed[h2*.h_number - beginHunk] := true;
		    h2 := nil;
		    symDelete(&h*.h_externs, en*.en_name, en*.en_nameLen);
		else
		    h2 := h2*.h_next;
		fi;
	    od;
	fi;
	en := en + sizeof(Entry_t);
    od;
corp;

proc processLibraries(bool define)bool:
    *File_t fl;
    register *Hunk_t h;
    register **Hunk_t hp;
    register *Entry_t en;
    register ulong count, hunkCount, beginHunk;
    ulong saveHunkNumber;
    bool definedNew, usedNew, foundIt;

    saveHunkNumber := HunkNumber;
    usedNew := true;
    FirstPass := true;
    while GlobalExtTable.tb_usedCount ~= 0 and usedNew do
	if DEBUG_LEVEL > 0 then
	    print("**** starting library scan\n");
	fi;
	HunkNumber := saveHunkNumber;
	usedNew := false;
	fl := LibraryList;
	while fl ~= nil do
	    if DEBUG_LEVEL > 0 then
		print("processing library '");
		print(fl*.fl_name);
		print("'\n");
	    fi;
	    setInput(fl*.fl_handle, fl*.fl_name);
	    fl*.fl_handle := 0;
	    FirstHunk := HunkNumber;
	    beginHunk := HunkNumber;
	    hp := &fl*.fl_hunkList;
	    hunkCount := 0;
	    while readRealHunk(hp, true, false) do
		h := hp*;
		hp := &h*.h_next;
		if hunkCount >= MAX_HUNKS then
		    abort("too many hunks");
		fi;
		HunksUsed[hunkCount] := false;
		/* have we already loaded this hunk? */
		if HunkMap*[h*.h_number] = nil then
		    /* does this hunk resolve an extref? */
		    en := h*.h_symbols.tb_entries;
		    count := h*.h_symbols.tb_entryCount;
		    while count ~= 0 do
			count := count - 1;
			if en*.en_nameLen ~= EN_EMPTY and
			    en*.en_nameLen ~= EN_FREE
			then
			    if symLookup(&GlobalExtTable,
					 en*.en_name, en*.en_nameLen) ~= nil
			    then
				HunksUsed[hunkCount] := true;
				usedNew := true;
				if DEBUG_LEVEL > 1 then
				    print("hunk ");
				    printNumber(h*.h_number);
				    print(" being selected - '");
				    print(en*.en_name);
				    print("'\n");
				fi;
				count := 0;
			    fi;
			fi;
			en := en + sizeof(Entry_t);
		    od;
		fi;
		hunkCount := hunkCount + 1;
	    od;
	    hp* := nil;

	    /* look for internal references, by extref or reloc */

	    definedNew := usedNew;
	    while definedNew do
		if DEBUG_LEVEL > 1 then
		    print("*** starting iterative scan of library\n");
		fi;
		definedNew := false;
		hp := &fl*.fl_hunkList;
		while
		    h := hp*;
		    h ~= nil
		do
		    if HunksUsed[h*.h_number - beginHunk] then
			definedNew := true;
			hp* := h*.h_next;
			selectHunk(fl, h, beginHunk);
			addHunkFromLibrary(fl, h, define);
		    else
			hp := &h*.h_next;
		    fi;
		od;
	    od;

	    /* remove unneeded hunks */
	    freeHunkList(fl*.fl_hunkList);
	    fl*.fl_hunkList := nil;

	    fl*.fl_handle := unsetInput();
	    fl := fl*.fl_next;
	od;
	FirstPass := false;
    od;
    not HadError
corp;

proc checkUnDef()bool:
    register *Entry_t en;
    register ulong count;

    if GlobalExtTable.tb_usedCount = 0 then
	return(true);
    fi;
    print("Undefined symbol(s):\n");
    count := GlobalExtTable.tb_entryCount;
    en := GlobalExtTable.tb_entries;
    while count ~= 0 do
	count := count - 1;
	if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
	    print("    '");
	    print(en*.en_name);
	    print("'\n");
	fi;
	en := en + sizeof(Entry_t);
    od;
    false
corp;

proc redoSymbolic(register *Hunk_t h)void:
    register *Symbol_t sy;

    sy := h*.h_symbolic;
    while sy ~= nil do
	sy*.sy_value := sy*.sy_value + h*.h_smallOffset;
	sy := sy*.sy_next;
    od;
corp;

proc readdressGlobals()void:
    register *Entry_t en;
    register ulong count;

    if DEBUG_LEVEL > 0 then
	print("readdressing symbols\n");
    fi;

    /* Bump the values of all symbols as required, since they will
       now be inside the single, merged hunk. */

    en := GlobalDefTable.tb_entries;
    count := GlobalDefTable.tb_entryCount;
    while count ~= 0 do
	count := count - 1;
	if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
	    en*.en_value :=
		en*.en_value + HunkMap*[en*.en_hunkNumber]*.h_smallOffset;
	fi;
	en := en + sizeof(Entry_t);
    od;
corp;

proc writeReloc(*Hunk_t h)void:
    register *Reloc_t rl;
    register ulong maxHunk, hunkNumber, totalCount;

    rl := h*.h_relocs;
    if rl = nil then
	return;
    fi;
    maxHunk := 0;
    while rl ~= nil do
	hunkNumber := HunkMap*[rl*.rl_hunkNumber]*.h_number;
	if hunkNumber > maxHunk then
	    maxHunk := hunkNumber;
	fi;
	rl := rl*.rl_next;
    od;

    writeLong(HUNK_RELOC32);

    hunkNumber := 0;
    while
	totalCount := 0;
	rl := h*.h_relocs;
	while rl ~= nil do
	    if rl*.rl_size = sz_32 then
		if HunkMap*[rl*.rl_hunkNumber]*.h_number = hunkNumber then
		    totalCount := totalCount + rl*.rl_count;
		fi;
	    fi;
	    rl := rl*.rl_next;
	od;
	if totalCount ~= 0 then
	    writeLong(totalCount);
	    writeLong(hunkNumber);
	    rl := h*.h_relocs;
	    while rl ~= nil do
		if rl*.rl_size = sz_32 then
		    if HunkMap*[rl*.rl_hunkNumber]*.h_number = hunkNumber then
			doWrite(rl*.rl_refOffsets, rl*.rl_count*sizeof(ulong));
		    fi;
		fi;
		rl := rl*.rl_next;
	    od;
	fi;
	hunkNumber ~= maxHunk
    do
	hunkNumber := hunkNumber + 1;
    od;

    writeLong(0);
corp;

proc writeSymbolic(*Hunk_t h; bool define)void:
    register *Symbol_t sy;
    register *Entry_t en @ sy;
    register ulong count;
    bool hadSymbolic;

    hadSymbolic := false;
    sy := h*.h_symbolic;
    if sy ~= nil then
	hadSymbolic := true;
	writeLong(HUNK_SYMBOL);
	while sy ~= nil do
	    if not define or
		symLookup(&h*.h_symbols, sy*.sy_name, sy*.sy_nameLen) = nil
	    then
		writeName(sy*.sy_name, sy*.sy_nameLen);
		writeLong(sy*.sy_value);
	    fi;
	    sy := sy*.sy_next;
	od;
    fi;
    if define and h*.h_symbols.tb_entryCount ~= 0 then
	count := h*.h_symbols.tb_entryCount;
	en := h*.h_symbols.tb_entries;
	while count ~= 0 do
	    count := count - 1;
	    if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
		if not hadSymbolic then
		    hadSymbolic := true;
		    writeLong(HUNK_SYMBOL);
		fi;
		writeName(en*.en_name, en*.en_nameLen);
		writeLong(en*.en_value + h*.h_smallOffset);
	    fi;
	    en := en + sizeof(Entry_t);
	od;
    fi;
    if hadSymbolic then
	writeLong(0);
    fi;
    if h*.h_debug ~= nil and h*.h_debugLen ~= 0 then
	writeLong(HUNK_DEBUG);
	writeLong(h*.h_debugLen / 4);
	doWrite(h*.h_debug, h*.h_debugLen);
    fi;

    writeLong(HUNK_END);
corp;

proc writeHunksOfType(register ulong hunkType; ulong size;
		      *Hunk_t firstHunk; bool define)void:
    register *Hunk_t h;
    register *Symbol_t sy;
    register *Entry_t en @ sy;
    register ulong count;
    bool hadSymbolic;

    writeLong(hunkType);
    writeLong(size / 4);
    h := firstHunk;
    while h ~= nil do
	if hunkType ~= HUNK_BSS and h*.h_type = hunkType and h*.h_length ~= 0
	then
	    doWrite(h*.h_buffer, h*.h_length);
	fi;
	h := h*.h_next;
    od;

    writeReloc(firstHunk);

    h := firstHunk;
    hadSymbolic := false;
    while h ~= nil do
	if h*.h_type = hunkType then
	    sy := h*.h_symbolic;
	    if sy ~= nil then
		if not hadSymbolic then
		    hadSymbolic := true;
		    writeLong(HUNK_SYMBOL);
		fi;
		while sy ~= nil do
		    if not define or
			symLookup(&h*.h_symbols, sy*.sy_name, sy*.sy_nameLen)
			    = nil
		    then
			writeName(sy*.sy_name, sy*.sy_nameLen);
			writeLong(sy*.sy_value);
		    fi;
		    sy := sy*.sy_next;
		od;
	    fi;
	fi;
	h := h*.h_next;
    od;
    if define then
	h := firstHunk;
	while h ~= nil do
	    if h*.h_type = hunkType then
		count := h*.h_symbols.tb_entryCount;
		en := h*.h_symbols.tb_entries;
		while count ~= 0 do
		    count := count - 1;
		    if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE
		    then
			if not hadSymbolic then
			    hadSymbolic := true;
			    writeLong(HUNK_SYMBOL);
			fi;
			writeName(en*.en_name, en*.en_nameLen);
			writeLong(en*.en_value + h*.h_smallOffset);
		    fi;
		    en := en + sizeof(Entry_t);
		od;
	    fi;
	    h := h*.h_next;
	od;
    fi;
    if hadSymbolic then
	writeLong(0);
    fi;

    h := firstHunk;
    while h ~= nil do
	if h*.h_type = hunkType and h*.h_debug ~= nil and h*.h_debugLen ~= 0
	then
	    writeLong(HUNK_DEBUG);
	    writeLong(h*.h_debugLen / 4);
	    doWrite(h*.h_debug, h*.h_debugLen);
	fi;
	h := h*.h_next;
    od;

    writeLong(HUNK_END);
corp;

proc fixRefs(*Ref_t rf; register *Hunk_t h)void:
    register *ulong up, usePtr32;
    register *uint usePtr16 @ usePtr32;
    register *ushort usePtr8 @ usePtr32;
    register ulong offset, count, length;
    register Size_t size;

    offset := rf*.rf_value*.en_value;
    count := rf*.rf_count;
    up := rf*.rf_offsets;
    size := rf*.rf_size;
    length := h*.h_length;
    while count ~= 0 do
	count := count - 1;
	/* bump the offset in the code/data by the offset of the
	   target symbol in its hunk */
	if up* + sizeof(ulong) > length then
	    print("DRLink: extern ref offset ");
	    printNumber(up*);
	    print(" > source hunk size ");
	    printNumber(h*.h_length);
	    abortTail();
	fi;
	usePtr32 := pretend(h*.h_buffer, *ulong) + up*;
	if size = sz_32 then
	    usePtr32* := usePtr32* + offset;
	elif size = sz_16 then
	    usePtr16* := usePtr16* + make(offset, uint);
	else
	    usePtr8* := usePtr8* + make(offset, ushort);
	fi;
	up* := up* + h*.h_smallOffset;
	up := up + sizeof(ulong);
    od;
corp;

proc fixRelocs(*Reloc_t rl; register *Hunk_t h; ulong hunkNumber)void:
    register *ulong up, usePtr32;
    register *uint usePtr16 @ usePtr32;
    register *ushort usePtr8 @ usePtr32;
    register ulong offset, count, length;
    register Size_t size;

    offset := HunkMap*[hunkNumber]*.h_smallOffset;
    up := rl*.rl_refOffsets;
    count := rl*.rl_count;
    size := rl*.rl_size;
    length := h*.h_length;
    while count ~= 0 do
	count := count - 1;
	if up* > length then
	    print("DRLink: reloc offset ");
	    printNumber(up*);
	    print(" > source hunk size ");
	    printNumber(length);
	    abortTail();
	fi;
	usePtr32 := pretend(h*.h_buffer, *ulong) + up*;
	if size = sz_32 then
	    usePtr32* := usePtr32* + offset;
	elif size = sz_16 then
	    usePtr16* := usePtr16* + make(offset, uint);
	else
	    usePtr8* := usePtr8* + make(offset, ushort);
	fi;
	up* := up* + h*.h_smallOffset;
	up := up + sizeof(ulong);
    od;
corp;

proc writeOutputFile(bool smallCode, smallData, smallBss, define)void:
    uint OP_JMP_ABS = 0x4ef9;
    register ulong count, hunkNumber, offset;
    register *Hunk_t h;
    register *Ref_t rf;
    register *Entry_t en;
    register *Reloc_t rl @ en, rl2 @ rf;
    register *uint wp @ rf;
    *Hunk_t kludgeHunk, h2, codeHunk, dataHunk, bssHunk;
    *ulong up, usePtr32;
    *uint usePtr16;
    *ushort usePtr8;
    ulong hunkCount, codeSize, dataSize, bssSize, otherSize;
    ulong firstCode, firstData, firstBss, specialHunk;
    bool seenCode, seenData, seenBss, special;

    if ForceSmallCode then
	smallCode := true;
    fi;
    if ForceSmallData then
	smallData := true;
    fi;
    if ForceSmallBss then
	smallBss := true;
    fi;

    h := HunkList;
    if h*.h_type = HUNK_CODE and h*.h_length = 0 and not smallCode then

	if DEBUG_LEVEL > 2 then
	    print("modifying first hunk to be a JMP\n");
	fi;

	/* gack! any other solution to this has too many other consequences */
	kludgeHunk := h;
	wp := getMem(8);	/* 6, rounded up to a multiple of 4 */
	h*.h_buffer := wp;
	wp* := OP_JMP_ABS;
	wp := wp + sizeof(uint);
	wp* := 0;
	wp := wp + sizeof(uint);
	wp* := 0;
	h*.h_length := 8;
	rl := getMem(sizeof(Reloc_t));
	rl*.rl_next := nil;
	rl*.rl_refOffsets := getMem(sizeof(ulong));
	rl*.rl_count := 1;
	rl*.rl_newCount := 1;
	rl*.rl_refOffsets* := 2;
	rl*.rl_size := sz_32;
	h*.h_relocs := rl;
    else
	kludgeHunk := nil;
    fi;

    if DEBUG_LEVEL > 0 then
	print("linking final refs to defs\n");
    fi;

    hunkNumber := 0;
    codeSize := 0;
    dataSize := 0;
    bssSize := 0;
    otherSize := 0;
    seenCode := false;
    seenData := false;
    seenBss := false;
    h := HunkList;
    while h ~= nil do
	/* Note that for smallXXX, we make all of a given type of hunk have
	   the output hunk number of the first one of that type, which will
	   be the only one of that type. This is so that 'writeReloc'
	   will get the relocs right. */
	case h*.h_type
	incase HUNK_CODE:
	    if smallCode then
		if not seenCode then
		    seenCode := true;
		    codeHunk := h;
		    firstCode := hunkNumber;
		    hunkNumber := hunkNumber + 1;
		fi;
		h*.h_number := firstCode;
		h*.h_smallOffset := codeSize;
		redoSymbolic(h);
	    else
		if h ~= kludgeHunk and kludgeHunk ~= nil then
		    kludgeHunk*.h_relocs*.rl_hunkNumber := hunkNumber;
		    kludgeHunk := nil;
		fi;
		h*.h_number := hunkNumber;
		h*.h_smallOffset := 0;
		hunkNumber := hunkNumber + 1;
	    fi;
	    codeSize := codeSize + h*.h_length;
	incase HUNK_DATA:
	    if smallData then
		if not seenData then
		    seenData := true;
		    dataHunk := h;
		    firstData := hunkNumber;
		    hunkNumber := hunkNumber + 1;
		fi;
		h*.h_number := firstData;
		h*.h_smallOffset := dataSize;
		redoSymbolic(h);
	    else
		h*.h_number := hunkNumber;
		h*.h_smallOffset := 0;
		hunkNumber := hunkNumber + 1;
	    fi;
	    dataSize := dataSize + h*.h_length;
	incase HUNK_BSS:
	    if smallBss then
		if not seenBss then
		    seenBss := true;
		    bssHunk := h;
		    firstBss := hunkNumber;
		    hunkNumber := hunkNumber + 1;
		fi;
		h*.h_number := firstBss;
		h*.h_smallOffset := bssSize;
		redoSymbolic(h);
	    else
		h*.h_number := hunkNumber;
		h*.h_smallOffset := 0;
		hunkNumber := hunkNumber + 1;
	    fi;
	    bssSize := bssSize + h*.h_length;
	default:
	    h*.h_number := hunkNumber;
	    h*.h_smallOffset := 0;
	    hunkNumber := hunkNumber + 1;
	    otherSize := otherSize + h*.h_length;
	esac;
	rf := h*.h_refs;
	while rf ~= nil do
	    en := symLookup(&GlobalDefTable, rf*.rf_name, rf*.rf_nameLen);
	    if en = nil then
		abortString("no definition for satisfied ref '",
			    rf*.rf_name, "'");
	    fi;
	    rf*.rf_value := en;
	    rf := rf*.rf_next;
	od;
	h := h*.h_next;
    od;
    hunkCount := hunkNumber;

    /* NOTE: At this point all hunks have had their h_number updated to be
       the number of the output hunk they will be or will be part of. All
       other hunk numbers (in relocs, refs, and GlobalDefTable) still use
       the original input hunk numbers. */

    if smallCode or smallData or smallBss then
	readdressGlobals();
    fi;

    if DEBUG_LEVEL > 0 then
	print("writing HUNK_HEADER\n");
    fi;

    writeLong(HUNK_HEADER);
    writeLong(0);
    writeLong(hunkNumber);
    writeLong(0);
    writeLong(hunkNumber - 1);
    seenCode := false;
    seenData := false;
    seenBss := false;
    h := HunkList;
    while h ~= nil do
	case h*.h_type
	incase HUNK_CODE:
	    if smallCode then
		if not seenCode then
		    seenCode := true;
		    writeLong(codeSize / 4);
		fi;
	    else
		writeLong(h*.h_length / 4);
	    fi;
	incase HUNK_DATA:
	    if smallData then
		if not seenData then
		    seenData := true;
		    writeLong(dataSize / 4);
		fi;
	    else
		writeLong(h*.h_length / 4);
	    fi;
	incase HUNK_BSS:
	    if smallBss then
		if not seenBss then
		    seenBss := true;
		    writeLong(bssSize / 4);
		fi;
	    else
		writeLong(h*.h_length / 4);
	    fi;
	default:
	    writeLong((h*.h_length / 4) | (h*.h_type & ~HUNK_MASK));
	esac;
	h := h*.h_next;
    od;

    if DEBUG_LEVEL > 0 then
	print("counting total refs\n");
    fi;

    /* Count the total refs needed to each other output hunk. Note that
       when 'smallXXX' is in effect, the relocs will all go on the first
       hunk of that type. 'rl_count' is the current size of the reloc
       arrays. We accumulate 'rl_newCount' as the new, desired size. */

    h := HunkList;
    while h ~= nil do
	case h*.h_type
	incase HUNK_CODE:
	    if smallCode then
		h2 := codeHunk;
	    else
		h2 := h;
	    fi;
	incase HUNK_DATA:
	    if smallData then
		h2 := dataHunk;
	    else
		h2 := h;
	    fi;
	incase HUNK_BSS:
	    if smallBss then
		h2 := bssHunk;
	    else
		h2 := h;
	    fi;
	default:
	    h2 := h;
	esac;

	rf := h*.h_refs;
	while rf ~= nil do
	    if rf*.rf_size = sz_32 then
		hunkNumber := rf*.rf_value*.en_hunkNumber;
		rl := h2*.h_relocs;
		while rl ~= nil and
		    (rl*.rl_hunkNumber ~= hunkNumber or rl*.rl_size ~= sz_32)
		do
		    rl := rl*.rl_next;
		od;
		if rl = nil then
		    rl := getMem(sizeof(Reloc_t));
		    rl*.rl_next := h2*.h_relocs;
		    rl*.rl_hunkNumber := hunkNumber;
		    rl*.rl_count := 0;
		    rl*.rl_newCount := rf*.rf_count;
		    rl*.rl_refOffsets := nil;
		    rl*.rl_size := sz_32;
		    h2*.h_relocs := rl;
		else
		    rl*.rl_newCount := rl*.rl_newCount + rf*.rf_count;
		fi;
	    fi;
	    rf := rf*.rf_next;
	od;

	if h ~= h2 then
	    /* SmallXXX - attach ALL relocs to this hunk. */
	    rl := h*.h_relocs;
	    while rl ~= nil do
		if rl*.rl_size = sz_32 then
		    hunkNumber := rl*.rl_hunkNumber;
		    rl2 := h2*.h_relocs;
		    while rl2 ~= nil and
			(rl2*.rl_hunkNumber ~= hunkNumber or
			 rl2*.rl_size ~= sz_32)
		    do
			rl2 := rl2*.rl_next;
		    od;
		    if rl2 = nil then
			rl2 := getMem(sizeof(Reloc_t));
			rl2*.rl_next := h2*.h_relocs;
			rl2*.rl_hunkNumber := hunkNumber;
			rl2*.rl_count := 0;
			rl2*.rl_newCount := rl*.rl_count;
			rl2*.rl_refOffsets := nil;
			rl2*.rl_size := sz_32;
			h2*.h_relocs := rl2;
		    else
			rl2*.rl_newCount := rl2*.rl_newCount + rl*.rl_count;
		    fi;
		fi;
		rl := rl*.rl_next;
	    od;
	fi;

	h := h*.h_next;
    od;

    /* allocate contiguous space for the refs in each reloc unit */

    if DEBUG_LEVEL > 0 then
	print("allocating total reloc space\n");
    fi;

    h := HunkList;
    while h ~= nil do
	rl := h*.h_relocs;
	while rl ~= nil do
	    fixRelocs(rl, h, rl*.rl_hunkNumber);
	    if rl*.rl_size = sz_32 then
		count := rl*.rl_count;
		if rl*.rl_newCount ~= count then
		    up := getMem(rl*.rl_newCount * sizeof(ulong));
		    if count ~= 0 then
			BlockCopy(up, rl*.rl_refOffsets,count * sizeof(ulong));
			FreeMem(rl*.rl_refOffsets, count * sizeof(ulong));
		    fi;
		    rl*.rl_refOffsets := up;
		    rl*.rl_count := rl*.rl_newCount;
		    rl*.rl_newCount := count;
		fi;
	    fi;
	    rl := rl*.rl_next;
	od;
	h := h*.h_next;
    od;

    /* Assemble all of the relocs together in the contiguous spaces. Note
       that this is the actual linking being done here! Note also that
       'rl_count' is now the size of the enlarged reloc array, and
       'rl_newCount' is used now as an accumulating offset. */

    if DEBUG_LEVEL > 0 then
	print("assembling total relocs\n");
    fi;

    h := HunkList;
    while h ~= nil do
	case h*.h_type
	incase HUNK_CODE:
	    if smallCode then
		h2 := codeHunk;
	    else
		h2 := h;
	    fi;
	incase HUNK_DATA:
	    if smallData then
		h2 := dataHunk;
	    else
		h2 := h;
	    fi;
	incase HUNK_BSS:
	    if smallBss then
		h2 := bssHunk;
	    else
		h2 := h;
	    fi;
	default:
	    h2 := h;
	esac;

	rf := h*.h_refs;
	while rf ~= nil do
	    fixRefs(rf, h);
	    if rf*.rf_size = sz_32 then
		hunkNumber := rf*.rf_value*.en_hunkNumber;
		rl := h2*.h_relocs;
		while
		    if rl = nil then
			abort("nil reloc when merging refs to relocs");
		    fi;
		    rl*.rl_hunkNumber ~= hunkNumber
		do
		    rl := rl*.rl_next;
		od;
		count := rf*.rf_count;
		BlockCopy(rl*.rl_refOffsets + rl*.rl_newCount * sizeof(ulong),
			  rf*.rf_offsets, count * sizeof(ulong));
		rl*.rl_newCount := rl*.rl_newCount + count;
	    fi;
	    rf := rf*.rf_next;
	od;

	if h ~= h2 then
	    /* this is not the final output hunk of this type - stick all of
	       my reloc info onto the end of the reloc info attached to the
	       final output hunk of this type */
	    rl := h*.h_relocs;
	    while rl ~= nil do
		if rl*.rl_size = sz_32 then
		    hunkNumber := rl*.rl_hunkNumber;
		    rl2 := h2*.h_relocs;
		    while rl2*.rl_hunkNumber ~= hunkNumber do
			rl2 := rl2*.rl_next;
			if rl2 = nil then
			    abort("nil reloc when merging relocs to relocs");
			fi;
		    od;
		    count := rl*.rl_count;
		    BlockCopy(
			rl2*.rl_refOffsets + rl2*.rl_newCount * sizeof(ulong),
			rl*.rl_refOffsets, count * sizeof(ulong));
		    rl2*.rl_newCount := rl2*.rl_newCount + count;
		fi;
		rl := rl*.rl_next;
	    od;
	fi;

	h := h*.h_next;
    od;

    if HunkList*.h_type ~= HUNK_CODE then
	print("*** Warning: first hunk is not a code hunk! ***\n");
    fi;

    /* Write the hunk data, if any. We have to do this AFTER the above
       reloc fiddling, since internal offsets are changed. */

    if DEBUG_LEVEL > 0 then
	print("outputting final hunks\n");
    fi;

    seenCode := false;
    seenData := false;
    seenBss := false;
    h := HunkList;
    while h ~= nil do
	special := false;
	case h*.h_type
	incase HUNK_CODE:
	    if smallCode then
		if not seenCode then
		    seenCode := true;
		    writeHunksOfType(HUNK_CODE, codeSize, h, define);
		fi;
		special := true;
	    fi;
	incase HUNK_DATA:
	    if smallData then
		if not seenData then
		    seenData := true;
		    writeHunksOfType(HUNK_DATA, dataSize, h, define);
		fi;
		special := true;
	    fi;
	incase HUNK_BSS:
	    if smallBss then
		if not seenBss then
		    seenBss := true;
		    writeHunksOfType(HUNK_BSS, bssSize, h, define);
		fi;
		special := true;
	    fi;
	esac;
	if not special then
	    writeLong(h*.h_type);
	    writeLong(h*.h_length / 4);
	    if h*.h_type & HUNK_MASK ~= HUNK_BSS and h*.h_length ~= 0 then
		doWrite(h*.h_buffer, h*.h_length);
	    fi;

	    /* and now write out the completed reloc information */

	    if DEBUG_LEVEL > 2 then
		print("writing total relocs\n");
	    fi;
	    writeReloc(h);
	    writeSymbolic(h, define);
	fi;
	h := h*.h_next;
    od;

    print("hunks: ");
    printNumber(hunkCount);
    print("  code: ");
    printNumber(codeSize);
    print("  data: ");
    printNumber(dataSize);
    print("  bss: ");
    printNumber(bssSize);
    print("  other: ");
    printNumber(otherSize);
    print("\n");
corp;
