#drinc:libraries/dos.g
#drinc:exec/memory.g
#DRLink.g

proc symCreate(register *Table_t tb; register ulong shift)void:
    register *Entry_t en;

    tb*.tb_usedCount := 0;
    tb*.tb_shift := shift;
    shift := 1 << shift;
    tb*.tb_entryCount := 0;
    tb*.tb_mask := shift - 1;
    en := getMem(sizeof(Entry_t) * shift);
    tb*.tb_entryCount := shift;
    tb*.tb_entries := en;
    while shift ~= 0 do
	shift := shift - 1;
	en*.en_nameLen := EN_EMPTY;
	en := en + sizeof(Entry_t);
    od;
corp;

proc symDestroy(*Table_t tb)void:
    register ulong n;
    register *Entry_t en;
    register proc(arbptr p; ulong len)void r_FreeMem;

    r_FreeMem := FreeMem;
    n := tb*.tb_entryCount;
    if n = 0 then
	return;
    fi;
    en := tb*.tb_entries;
    while n ~= 0 do
	n := n - 1;
	if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE and
	    en*.en_name ~= nil
	then
	    r_FreeMem(en*.en_name, en*.en_nameLen);
	fi;
	en := en + sizeof(Entry_t);
    od;
    r_FreeMem(tb*.tb_entries, tb*.tb_entryCount * sizeof(Entry_t));
    tb*.tb_entryCount := 0;
corp;

proc symFlush(register *Table_t tb)void:
    register ulong n;
    register *Entry_t en;

    n := tb*.tb_entryCount;
    en := tb*.tb_entries;
    while n ~= 0 do
	n := n - 1;
	en*.en_nameLen := EN_EMPTY;
	en := en + sizeof(Entry_t);
    od;
    tb*.tb_usedCount := 0;
corp;

proc symEmpty(register *Table_t tb)void:
    register ulong n;
    register *Entry_t en;

    n := tb*.tb_entryCount;
    en := tb*.tb_entries;
    while n ~= 0 do
	n := n - 1;
	if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE and
	    en*.en_name ~= nil
	then
	    FreeMem(en*.en_name, en*.en_nameLen);
	fi;
	en*.en_nameLen := EN_EMPTY;
	en := en + sizeof(Entry_t);
    od;
    tb*.tb_usedCount := 0;
corp;

proc symSearch(*Table_t tb; *char name; uint nameLen)*Entry_t:
    register *Entry_t en;
    *Entry_t found, enEnd, start;
    register *char p, q;
    register ulong hash, tmp;
    register uint len;
    bool first;

    p := name;
    len := nameLen;
    hash := p* - '\e';
    p := p + sizeof(char);
    if len > 1 then
	hash := hash << 8;
	tmp := p* - '\e';
	hash := hash | tmp;
	if len > 2 then
	    hash := hash << 8;
	    p := p + len * sizeof(char);
	    p := p - 3 * sizeof(char);
	    tmp := p* - '\e';
	    hash := hash | tmp;
	    if len > 3 then
		hash := hash << 8;
		p := p - sizeof(char);
		tmp := p* - '\e';
		hash := hash | tmp;
	    fi;
	fi;
    fi;
    tmp := hash;
    tmp := tmp & tb*.tb_mask;
    en := tb*.tb_entries;
    enEnd := en + sizeof(Entry_t) << tb*.tb_shift - sizeof(Entry_t);
    for len from sizeof(Entry_t) - 1 downto 0 do
	en := en + tmp;
    od;
    found := nil;
    start := en;
    first := true;
    while
	if en = start and not first then
	    /* scanned entire table - name not found - use the first of the
	       free slots we passed. */
	    en := found;
	    false
	elif en*.en_nameLen = EN_EMPTY then
	    /* found an unused slot - name is not in table */
	    false
	elif en*.en_nameLen = EN_FREE then
	    /* found a slot that used to be used - have to keep looking, but
	       remember this slot for entry if it is the first of them */
	    if found = nil then
		found := en;
	    fi;
	    true
	elif en*.en_hash ~= hash or en*.en_nameLen ~= nameLen then
	    /* match not possible */
	    true
	else
	    /* have to actually compare strings */
	    p := name;
	    q := en*.en_name;
	    len := nameLen;
	    while len ~= 0 and p* = q* do
		len := len - 1;
		p := p + sizeof(char);
		q := q + sizeof(char);
	    od;
	    len ~= 0
	fi
    do
	if en = enEnd then
	    en := tb*.tb_entries;
	else
	    en := en + sizeof(Entry_t);
	fi;
	first := false;
    od;
    if en*.en_nameLen = EN_EMPTY or en*.en_nameLen = EN_FREE then
	if found ~= nil then
	    en := found;
	fi;
	en*.en_hash := hash;
    fi;
    en
corp;

proc symTransfer(*Table_t oldTable; register *Table_t newTable;
		 ulong shift)void:
    register *Entry_t en;
    register ulong count;

    symCreate(newTable, shift);
    newTable*.tb_usedCount := oldTable*.tb_usedCount;
    count := oldTable*.tb_entryCount;
    en := oldTable*.tb_entries;
    while count ~= 0 do
	count := count - 1;
	if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
	    symSearch(newTable, en*.en_name, en*.en_nameLen)* := en*;
	fi;
	en := en + sizeof(Entry_t);
    od;
corp;

proc symEnter(*Table_t tb; *char name; uint nameLen)*Entry_t:
    register *Entry_t en;
    register *char p, q;
    Table_t newTable;
    register uint len;

    if tb*.tb_usedCount >= tb*.tb_entryCount * 4 / 5 then
	symTransfer(tb, &newTable, tb*.tb_shift + 1);
	FreeMem(tb*.tb_entries, tb*.tb_entryCount * sizeof(Entry_t));
	tb* := newTable;
    fi;
    en := symSearch(tb, name, nameLen);
    if en*.en_nameLen = EN_EMPTY or en*.en_nameLen = EN_FREE then
	en*.en_name := nil;
	en*.en_name := getMem(nameLen);
	en*.en_nameLen := nameLen;
	p := name;
	q := en*.en_name;
	len := nameLen;
	while len ~= 0 do
	    len := len - 1;
	    q* := p*;
	    q := q + sizeof(char);
	    p := p + sizeof(char);
	od;
	tb*.tb_usedCount := tb*.tb_usedCount + 1;
    fi;
    en
corp;

proc symLookup(register *Table_t tb; *char name; uint nameLen)*Entry_t:
    register *Entry_t en;

    if tb*.tb_usedCount = 0 or tb*.tb_entryCount = 0 then
	return(nil);
    fi;
    en := symSearch(tb, name, nameLen);
    if en*.en_nameLen = EN_EMPTY or en*.en_nameLen = EN_FREE then
	en := nil;
    fi;
    en
corp;

proc symDelete(register *Table_t tb; *char name; uint nameLen)void:
    register *Entry_t en;

    if tb*.tb_usedCount = 0 or tb*.tb_entryCount = 0 then
	return;
    fi;
    en := symSearch(tb, name, nameLen);
    if en*.en_nameLen ~= EN_EMPTY and en*.en_nameLen ~= EN_FREE then
	FreeMem(en*.en_name, en*.en_nameLen);
	en*.en_nameLen := EN_FREE;
	tb*.tb_usedCount := tb*.tb_usedCount - 1;
    fi;
corp;

proc symCopy(*Table_t oldTable, newTable)void:
    register ulong count, shift, capacity;

    count := oldTable*.tb_usedCount;
    if count = 0 then
	newTable*.tb_entryCount := 0;
	return;
    fi;
    shift := 1;
    capacity := 2;
    while capacity < count do
	capacity := capacity * 2;
	shift := shift + 1;
    od;
    symTransfer(oldTable, newTable, shift);
corp;
